@@ -6,6 +6,8 @@ from typing import Union as _Union
 import numpy as _np
 import pandas as _pd
 
+from . import gn_datetime as _gn_datetime
+
 
 def update_mindex(dataframe, lvl_name, loc=0, axis=1):
     """Inserts a level named as lvl_name into dataframe df in loc position.
@@ -244,4 +246,8 @@ def decdeg2degminsec(a: _np.ndarray):
 
 def throw_if_index_duplicates(df):
     if df.index.has_duplicates:
-        raise ValueError(f"Duplicated index entries present, see below:\n {df[df.index.duplicated()].to_string()}")
+        df_dupl = df[df.index.duplicated()]
+        dt_dupl = _gn_datetime.j20002datetime(df_dupl.index.get_level_values("TIME").unique().values)
+        raise ValueError(
+            f"Found duplicated index entries at {dt_dupl} time values. Complete duplicated index entries present, see below:\n {df[df.index.duplicated()].to_string()}"
+        )
