@@ -17,12 +17,47 @@ _RE_BLK_HEAD = _re.compile(rb"\+\w+\/\w+(\s[LU]|)\s*(CORR|COVA|INFO|)[ ]*\n(?:\*
 _RE_STATISTICS = _re.compile(r"^[ ]([A-Z (-]+[A-Z)])[ ]+([\d+\.]+)", _re.MULTILINE)
 
 
-def snx_soln_converter(soln):
-    try:
-        return int(soln)
-    except ValueError:
-        _logging.debug(f"Converting non-integer solution marker ({soln}) to 0")
-        return 0
+def get_snx_header(path_or_bytes):
+    # read line that starts with %=SNX and parse it into dict
+    snx_bytes = _gn_io.common.path2bytes(path_or_bytes)
+    header_begin = snx_bytes.find(b"%=SNX")
+    header_end = snx_bytes.find(b"\n", header_begin)
+    snx_hline = snx_bytes[header_begin:header_end].decode()
+    dates = _gn_datetime.yydoysec2datetime([snx_hline[15:27], snx_hline[32:44], snx_hline[45:57]]).tolist()
+    return {
+        "version": float(snx_hline[6:10]),
+        "ac_created": snx_hline[11:14],  # agency creating the file
+        "creation_time": dates[0],  # creation time of this SINEX file
+        "ac_data_prov": snx_hline[28:31],  # agency providing the data in the SINEX file
+        "data_start": dates[1],  # start time of the data used in the SINEX solution
+        "data_end": dates[2],  # end time of the data used in the SINEX solution
+        "obs_code": snx_hline[58],  # technique used to generate the SINEX solution
+        "n_estimates": int(snx_hline[60:65]),  # number of parameters estimated in this SINEX file
+        "constr_code": snx_hline[66],  # constraint in the SINEX solution
+        "content": tuple(snx_hline[68:79:2]),  # solution types contained in this SINEX file
+    }
+
+
+def format_snx_header(h: dict):
+    dates_yydoysec = _gn_datetime.j20002yydoysec(
+        _np.asarray([h["creation_time"], h["data_start"], h["data_end"]])
+    ).tolist()
+    # we override the format version with 2.02
+    line = f"%=SNX 2.02 {h['ac_created']} {dates_yydoysec[0]} {h['ac_data_prov']} {dates_yydoysec[1]} {dates_yydoysec[2]} {h['obs_code']} {h['n_estimates']:05} {h['constr_code']:1} {' '.join(h['content'])}"
+    return line
+
+
+def snx_soln_str_to_int(soln: _pd.Series) -> _pd.Series:
+    """Converts sinex solutions series to int with all non-numeric values, e.g. `----`, swapped with zeros"""
+    soln_int = _pd.to_numeric(soln, errors="coerce").fillna(0).astype(int)
+    return soln_int
+
+
+def snx_soln_int_to_str(soln: _pd.Series, nan_as_dash=True) -> _pd.Series:
+    """Converts int sinex solutions series back to str, swapping zeros with '----', as zeros are not part of the format"""
+    soln_str = soln.astype(str)
+    soln_str[~soln.values.astype(bool)] = "----" if nan_as_dash else ""  # use zeros as False
+    return soln_str
 
 
 def _get_valid_stypes(stypes):
@@ -105,7 +140,6 @@ def _snx_extract(snx_bytes, stypes, obj_type, verbose=True):
                 objects_in_buf += 1
             else:
                 _logging.error(f"{stype} ({stypes_dict[stype]}) blk not found")
-                # return None
                 objects_in_buf += 1
 
         else:
@@ -189,11 +223,10 @@ def snxdf2xyzdf(snxdf, unstack=True, keep_all_soln=None):
         ]
     ).values
     snxdf.drop(index=snxdf.index.values[~types_mask], inplace=True)
-    snxdf["CODE_PT"] = snxdf.CODE.values + "_" + snxdf.PT.values.astype(object)
-    snx_df = snxdf.drop(columns=["CODE", "PT"]).set_index(["CODE_PT", "REF_EPOCH", "TYPE"])
+    snx_df = snxdf.set_index(["CODE_PT", "REF_EPOCH", "TYPE"])
     if keep_all_soln is None:
-        # Retain existing behaviour by (all solutions but no solution number) default
-        snx_df = snx_df.drop(columns=["SOLN"])
+        # Retain existing behaviour by (all solutions but no solution number) default. Specifying level stops performance warning
+        snx_df = snx_df.drop(columns=["SOLN"], level=0)
     elif keep_all_soln:
         # Keep all solutions but include SOLN as an index value, not a column to avoid reshape issues
         snx_df = snx_df.set_index("SOLN", append=True)
@@ -206,16 +239,25 @@ def snxdf2xyzdf(snxdf, unstack=True, keep_all_soln=None):
     return snx_df.unstack(2) if unstack else snx_df
 
 
-def _get_snx_vector(path_or_bytes, stypes=("EST", "APR"), snx_format=True, keep_all_soln=None, verbose=True):
+def _get_snx_vector(
+    path_or_bytes,
+    stypes=("EST", "APR"),
+    format="long",
+    keep_all_soln=None,
+    verbose=True,
+    recenter_epochs=False,
+):
     """stypes = "APR","EST","NEQ"
     APRIORY, ESTIMATE, NORMAL_EQUATION
+
+    recenter_epochs would override the read-in time values with _gn_const.SEC_IN_DAY // 2 so same-day values from different sinex files will align. The actual timestamps could be close to 43200 but not exactly. This switch insures the alignment
     """
     path = None
     if isinstance(path_or_bytes, str):
         path = path_or_bytes
         snx_bytes = _gn_io.common.path2bytes(path)
     elif isinstance(path_or_bytes, list):
-        path, stypes, snx_format, verbose = path_or_bytes
+        path, stypes, format, verbose = path_or_bytes
         snx_bytes = _gn_io.common.path2bytes(path)
     else:
         snx_bytes = path_or_bytes
@@ -234,6 +276,7 @@ def _get_snx_vector(path_or_bytes, stypes=("EST", "APR"), snx_format=True, keep_
     if extracted is None:
         return None
     snx_buffer, stypes_rows, stypes_form, stypes_content = extracted
+    stypes = stypes_rows.keys()
 
     try:
         vector_raw = _pd.read_csv(
@@ -241,19 +284,19 @@ def _get_snx_vector(path_or_bytes, stypes=("EST", "APR"), snx_format=True, keep_
             delim_whitespace=True,
             comment=b"*",
             header=None,
-            usecols=[0, 1, 2, 3, 4, 5, 8, 9],
-            names=["INDEX", "TYPE", "CODE", "PT", "SOLN", "REF_EPOCH", "EST", "STD"],
+            usecols=[1, 2, 3, 4, 5, 6, 7, 8, 9],
+            names=["TYPE", "CODE", "PT", "SOLN", "REF_EPOCH", "UNIT", "CONSTR", "EST", "STD"],
             dtype={
-                "INDEX": int,
-                "TYPE": _gn_const.TYPE_CATEGORY,
+                "TYPE": object,
                 "CODE": object,
-                "PT": _gn_const.PT_CATEGORY,
+                "PT": object,
+                "SOLN": object,
                 "REF_EPOCH": object,
+                "UNIT": _gn_const.UNIT_CATEGORY,
+                "CONSTR": int,
                 "EST": _np.float_,
                 "STD": _np.float_,
             },
-            converters={"SOLN": snx_soln_converter},
-            index_col="INDEX",
         )
 
     except ValueError as _e:
@@ -263,30 +306,31 @@ def _get_snx_vector(path_or_bytes, stypes=("EST", "APR"), snx_format=True, keep_
         else:
             raise _e
 
-    if path is not None:
-        del snx_buffer  # need to test this better
+    stypes_vals = (
+        vector_raw.iloc[:, -2:].values.reshape(len(stypes), -1, 2).T
+    )  # converting the consecutively structured VALs and STDs of stypes blocks into vertically stacked arrays of VAls-STDs pairs for each stype
+    df = vector_raw.iloc[: stypes_vals.shape[1], :5]
+    df.attrs["aux"] = vector_raw.iloc[: stypes_vals.shape[1], [5, 6]]
 
-    vector_raw.index = vector_raw.index.values - 1  # start with 0
+    df.REF_EPOCH = _gn_datetime.yydoysec2datetime(df.REF_EPOCH, recenter=recenter_epochs, as_j2000=True)
+    df.SOLN = snx_soln_str_to_int(df.SOLN)
+    df["CODE_PT"] = df.CODE.values + "_" + df.PT.values
+    df.drop(["CODE", "PT"], axis=1, inplace=True)
 
-    output = []
-    prev_idx = 0
-    for i in range(len(stypes_rows)):
-        stype = stypes[i]
-        idx = stypes_rows[stype]
-        vec_df = (vector_raw[prev_idx : prev_idx + idx]).copy()
-        if i == 0:
-            vec_df.REF_EPOCH = _gn_datetime.yydoysec2datetime(vec_df.REF_EPOCH, recenter=True, as_j2000=True)
-        else:
-            vec_df = vec_df.iloc[:, 5:]
-            if vec_df.shape[0] != n_header:
-                vec_df = vec_df.reindex(_np.arange(start=0, stop=n_header), fill_value=0)
-        if stype in ["APR", "NEQ"]:
-            vec_df.rename(columns={"EST": stype}, inplace=True)
-            vec_df.drop(columns="STD", inplace=True)
-        output.append(vec_df)
-        prev_idx += idx
-    output = _pd.concat(output, axis=1)
-    return snxdf2xyzdf(output, unstack=(not snx_format), keep_all_soln=keep_all_soln)
+    m1 = _pd.MultiIndex.from_product([df.columns, [""]])
+    m2 = _pd.MultiIndex.from_product([["VAL", "STD"], stypes])
+
+    df[list(range(len(stypes) * 2))] = _np.hstack(stypes_vals)
+
+    df.columns = m1.append(m2)
+    if format == "wide":
+        return snxdf2xyzdf(df, unstack=True, keep_all_soln=keep_all_soln)
+    elif format == "long":
+        return snxdf2xyzdf(df, unstack=False, keep_all_soln=keep_all_soln)
+    elif format == "raw":
+        return df
+    else:
+        raise NotImplementedError
 
 
 def _matrix_raw2square(matrix_raw, stypes_form, n_elements=None):
@@ -355,11 +399,17 @@ def unc_snx(path, snx_format=True):
     return snxdf2xyzdf(output)
 
 
-def _read_snx_solution(path_or_bytes):
+def _read_snx_solution(path_or_bytes, recenter_epochs=False):
     """_get_snx_vector template to get a df with multiIndex columns as:
     | APR | EST | STD |
     |X|Y|Z|X|Y|Z|X|Y|Z|"""
-    return _get_snx_vector(path_or_bytes=path_or_bytes, stypes=("APR", "EST"), snx_format=False, verbose=False)
+    return _get_snx_vector(
+        path_or_bytes=path_or_bytes,
+        stypes=("APR", "EST"),
+        format="wide",
+        verbose=False,
+        recenter_epochs=recenter_epochs,
+    )
 
 
 # TODO get rid of p_tqdm. Need to rewrite hte loop with multiprocessing Pool
@@ -379,7 +429,7 @@ def _read_snx_solution(path_or_bytes):
 #     # return _pd.concat(data, axis=0).pivot(index=['CODE','TYPE'],columns='REF_EPOCH').T
 
 
-def _get_snx_vector_gzchunks(filename, block_name="SOLUTION/ESTIMATE", size_lookback=100):
+def _get_snx_vector_gzchunks(filename, block_name="SOLUTION/ESTIMATE", size_lookback=100, format="raw"):
     """extract block from a large gzipped sinex file e.g. ITRF2014 sinex"""
     block_open = False
     block_bytes = b""
@@ -412,19 +462,22 @@ def _get_snx_vector_gzchunks(filename, block_name="SOLUTION/ESTIMATE", size_look
                     stop = True
             i += 1
 
-    return _get_snx_vector(path_or_bytes=block_bytes, stypes=["EST"], snx_format=None)
+    return _get_snx_vector(path_or_bytes=block_bytes, stypes=["EST"], format=format)
 
 
-# SINEX ID BLOCK
 def _get_snx_id(path):
+    """
+    SINEX observation codes: C-Combined techniques used, D-DORIS, L-SLR, M-LLR, P-GNSS, R-VLBI.
+    """
     snx_bytes = _gn_io.common.path2bytes(path)
     site_id = _snx_extract_blk(snx_bytes=snx_bytes, blk_name="SITE/ID", remove_header=True)[0]
 
     id_df = _pd.read_fwf(
-        _BytesIO(site_id), header=None, colspecs=[(0, 5), (5, 8), (8, 18), (18, 20), (44, 55), (55, 68), (68, 76)]
+        _BytesIO(site_id),
+        header=None,
+        colspecs=[(0, 5), (5, 8), (8, 18), (18, 20), (20, 44), (44, 55), (55, 68), (68, 76)],
+        names=["CODE", "PT", "DOMES", "OBS_CODE", "DESCRIPTION", "LON", "LAT", "H"],
     )
-
-    id_df.columns = ["CODE", "PT", "DOMES", "T", "LON", "LAT", "H"]  # location may have non-unicode chars
     id_df.LON = _gn_aux.degminsec2decdeg(id_df.LON)
     id_df.LAT = _gn_aux.degminsec2decdeg(id_df.LAT)
     return id_df
@@ -520,3 +573,164 @@ def logllh2snxdms(llh):
     out[step1_mask] = buf
     out[~step1_mask] = " 000 00 00.0  00 00 00.0   000.0"
     return out.astype(object)
+
+
+def snx_format_float(float_series, format_string=r"{:21.14E}", all_positive=False):
+    """
+    This function can replicate fortran-like formatting that CODE uses in its sinex vector blocks, though absolutely unnecessary
+    ESTIMATE block values format: r"{:21.14E}"
+    standard deviation format: r"{:12.5E}" and all_positive=True
+    """
+    raw_str = float_series.apply(format_string.format)
+    negative_mask = float_series < 0
+
+    ndarr = raw_str.values.astype(str).view(("U1")).reshape(raw_str.shape[0], -1).astype(object)
+    exp = (ndarr[:, -3:]).astype(object).sum(axis=1).astype(int) + 1
+
+    negative_exp = exp < 0
+    ndarr[negative_exp, -3] = "-"
+    ndarr[~negative_exp, -3] = "+"
+
+    ndarr[:, -2:] = _pd.Series(_np.abs(exp).astype(str)).str.zfill(2).values.astype(str).view("U1").reshape(-1, 2)
+    ndarr[:, 1:3] = _np.roll(ndarr[:, 1:3], 1, 1)
+    if all_positive:
+        return ndarr[:, 1:].sum(axis=1)
+    ndarr[~negative_mask, 0] = "0"
+    return ndarr.sum(axis=1)  # concatanate all U1 strings together to get formatted strings
+
+
+def format_snx_matrix(matrices: _np.ndarray, matrices_types: dict, zeros_to_spaces=True) -> list:
+    """Vectorized matrix formatter that takes in any number of matrices and outputs consequtive formatted lists"""
+    matrix = matrices[0]
+    axis_len = matrix.shape[1]
+    div_3_size = (
+        _np.math.ceil(axis_len / 3) * 3
+    )  # we should be able to reshape into 3 cols -> this is a 3-divisible axis size
+
+    m_buf = []
+    for m in matrices:  # sizes must be equal
+        m_buf.append(
+            _np.hstack([_np.tril(m), _np.zeros(shape=(axis_len, div_3_size - axis_len))]).reshape((axis_len, -1, 3))
+        )
+    m_ready_to_reshape = _np.vstack(m_buf)
+    # adding elements to rows, so we could reshape each row into a set of 3-elements
+
+    # all_zero_mask = _np.all(m_ready_to_reshape == 0, axis=2)  # all true (equal 0)
+    index1 = _np.broadcast_to(_np.arange(1, axis_len + 1)[:, None], (axis_len, m_ready_to_reshape.shape[1])).reshape(
+        (-1)
+    )  # row
+    index2 = _np.broadcast_to(_np.arange(1, div_3_size, 3)[None, :], (axis_len, m_ready_to_reshape.shape[1])).reshape(
+        (-1)
+    )  # col
+    blk_length = index1.shape[0]
+    index = (
+        _pd.Series(_np.hstack([index1, index2]).astype(str)).str.rjust(6).values.reshape((-1, 2), order="F").sum(axis=1)
+    )
+    # add header field to index
+    index = _np.tile(index, len(matrices))
+
+    blk_bounds = {
+        "APR": f"SOLUTION/MATRIX_APRIORI L {matrices_types['APR'] if 'APR' in matrices_types.keys() else ''}\n",
+        "EST": f"SOLUTION/MATRIX_ESTIMATE L {matrices_types['EST'] if 'EST' in matrices_types.keys() else ''}\n",
+        "NEQ": f"SOLUTION/NORMAL_EQUATION_MATRIX L\n",
+    }
+    blk_header = "*PARA1 PARA2 ____PARA2+0__________ ____PARA2+1__________ ____PARA2+2__________\n"
+
+    keys = list(matrices_types.keys())
+
+    for i in range(len(keys)):
+        index[:-1:blk_length][i] = "+" + blk_bounds[keys[i]] + blk_header + index[:-1:blk_length][i]
+        if i > 0:
+            index[:-1:blk_length][i] = "-" + blk_bounds[keys[i]] + index[:-1:blk_length][i]
+
+    df = _pd.DataFrame(data=m_ready_to_reshape.reshape((-1, 3)))
+
+    non_zero_mask = ~_np.all(df.values == 0, axis=1)
+    # make sure that first lines of every matrix are not zero
+    non_zero_mask[:-1:blk_length] = True
+
+    c = df[non_zero_mask].applymap("{:21.14e}".format).values
+    if zeros_to_spaces:
+        plus2_mask = df.values[non_zero_mask, 2] == 0
+        plus3_mask = df.values[non_zero_mask, 1] == 0
+        c[plus2_mask, 2] = " " * 22
+        c[plus2_mask & plus3_mask, 1] = " " * 22
+        # how to write diagonal in this case??? chess order values may be hard to read with read_csv
+
+    return (index[non_zero_mask] + " " + c[:, 0] + " " + c[:, 1] + " " + c[:, 2]).tolist() + [
+        "-" + blk_bounds[keys[-1]]
+    ]
+
+
+def format_snx_vector(vector_df: _pd.DataFrame) -> list:
+    df = vector_df.reset_index()
+
+    VAL = df.VAL.stack().apply(r"{:21.14e}".format)
+
+    STD = df.STD.stack(dropna=False).apply(r"{:11.5e}".format)  # standard has 11.6e, however IGS uses 11.5e
+    STD[STD == "        nan"] = "           "
+
+    CODE_PT = df.CODE_PT.str.split("_", expand=True, n=2)
+
+    blk_common = (
+        _pd.Series(list(range(1, df.shape[0] + 1))).astype(str).str.rjust(6)
+        + " "
+        + df.TYPE.str.ljust(6)
+        + " "
+        + CODE_PT.iloc[:, 0].str.ljust(4)
+        + " "
+        + CODE_PT.iloc[:, 1].str.rjust(2)
+        + " "
+        + snx_soln_int_to_str(df.SOLN, nan_as_dash=True).str.rjust(4)
+        + " "
+        + _gn_datetime.datetime2yydoysec(_gn_datetime.j20002datetime(df.REF_EPOCH.values))
+        + " "
+        + df.attrs["aux"].UNIT.str.ljust(4).values
+        + " "
+        + df.attrs["aux"].CONSTR.values.astype("U1")
+    )
+
+    blk_bounds = _pd.Series(
+        index=["APR", "EST", "NEQ"],
+        data=["SOLUTION/APRIORI\n", "SOLUTION/ESTIMATE\n", "SOLUTION/NORMAL_EQUATION_VECTOR\n"],
+    )
+
+    blk_header = _pd.Series(
+        index=["APR", "EST", "NEQ"],
+        data=[
+            "*INDEX TYPE__ CODE PT SOLN _REF_EPOCH__ UNIT S __APRIORI VALUE______ _STD_DEV___\n",
+            "*INDEX TYPE__ CODE PT SOLN _REF_EPOCH__ UNIT S __APRIORI VALUE______ _STD_DEV___\n",
+            "*INDEX TYPE__ CODE PT SOLN _REF_EPOCH__ UNIT S __RIGHT_HAND_SIDE____\n",
+        ],
+    )
+    n_blks = df.columns.levels[1].size - 1  # n blks present
+    blk_series = _np.repeat(blk_common.values, n_blks) + " " + VAL + " " + STD
+
+    idx_first, idx_last = 0, blk_series.index.levels[0][-1]
+
+    blk_series[idx_first].update("+" + blk_bounds + blk_header + blk_series[idx_first])
+    blk_series[idx_last].update(blk_series[idx_last] + "\n-" + blk_bounds)
+
+    return blk_series.sort_index(level=1).to_list()
+
+
+def format_snx_id(snx_id):
+    lonlat = _gn_aux.decdeg2degminsec(snx_id[["LON", "LAT"]].values)
+    blk_series = (
+        snx_id.CODE.str.rjust(5)
+        + snx_id.PT.str.rjust(3)
+        + snx_id.DOMES.str.rjust(10)
+        + " "
+        + snx_id.DESCRIPTION.str.ljust(23)
+        + lonlat[:, 0]
+        + " "
+        + lonlat[:, 1]
+        + snx_id.H.round(1).astype(str).str.rjust(8)
+    )
+
+    return (
+        ["+SITE/ID\n"]
+        + ["*CODE PT __DOMES__ T _STATION DESCRIPTION__ APPROX_LON_ APPROX_LAT_ _APP_H_\n"]
+        + blk_series.to_list()
+        + ["-SITE/ID\n"]
+    )
